using System;
using System.IO;
using System.Text;
using System.Reflection;
using System.Collections;
using Server;

namespace Server.Scripts.Commands
{
	public class Docs
	{
		public static void Initialize()
		{
			Server.Commands.Register( "DocGen", AccessLevel.Administrator, new CommandEventHandler( DocGen_OnCommand ) );
		}

		private static void DocGen_OnCommand( CommandEventArgs e )
		{
			World.Broadcast( 0x35, true, "Documentation is being generated, please wait." );

			DateTime startTime = DateTime.Now;

			Document();

			DateTime endTime = DateTime.Now;

			World.Broadcast( 0x35, true, "Documentation has been completed. The entire process took {0:F1} seconds.", (endTime - startTime).TotalSeconds );
		}

		private class MemberComparer : IComparer
		{
			public int Compare( object x, object y )
			{
				if ( x == y )
					return 0;

				ConstructorInfo aCtor = x as ConstructorInfo;
				ConstructorInfo bCtor = y as ConstructorInfo;

				PropertyInfo aProp = x as PropertyInfo;
				PropertyInfo bProp = y as PropertyInfo;

				MethodInfo aMethod = x as MethodInfo;
				MethodInfo bMethod = y as MethodInfo;

				int v = 0;

				if ( aCtor != null )
				{
					if ( bCtor == null )
						v = -1;
				}
				else if ( bCtor != null )
				{
					if ( aCtor == null )
						v = 1;
				}
				else if ( aProp != null )
				{
					if ( bProp == null )
						v = -1;
				}
				else if ( bProp != null )
				{
					if ( aProp == null )
						v = 1;
				}

				if ( v == 0 )
				{
					v = GetNameFrom( aCtor, aProp, aMethod ).CompareTo( GetNameFrom( bCtor, bProp, bMethod ) );
				}

				if ( v == 0 && aCtor != null && bCtor != null )
				{
					v = aCtor.GetParameters().Length.CompareTo( bCtor.GetParameters().Length );
				}
				else if ( v == 0 && aMethod != null && bMethod != null )
				{
					v = aMethod.GetParameters().Length.CompareTo( bMethod.GetParameters().Length );
				}

				return v;
			}

			private string GetNameFrom( ConstructorInfo ctor, PropertyInfo prop, MethodInfo method )
			{
				if ( ctor != null )
					return ctor.DeclaringType.Name;
				else if ( prop != null )
					return prop.Name;
				else if ( method != null )
					return method.Name;
				else
					return "";
			}
		}

		private class TypeComparer : IComparer
		{
			public int Compare( object x, object y )
			{
				Type a = x as Type;
				Type b = y as Type;

				return a.Name.CompareTo( b.Name );
			}
		}

		private class NamespaceComparer : IComparer
		{
			public int Compare( object x, object y )
			{
				DictionaryEntry a = (DictionaryEntry)x;
				DictionaryEntry b = (DictionaryEntry)y;

				return ((string)a.Key).CompareTo( (string)b.Key );
			}
		}

		private static string[,] m_Aliases = new string[,]
			{
				{ "System.Object",	"<font color=\"blue\">object</font>" },
				{ "System.String",	"<font color=\"blue\">string</font>" },
				{ "System.Boolean",	"<font color=\"blue\">bool</font>" },
				{ "System.Byte",	"<font color=\"blue\">byte</font>" },
				{ "System.SByte",	"<font color=\"blue\">sbyte</font>" },
				{ "System.Int16",	"<font color=\"blue\">short</font>" },
				{ "System.UInt16",	"<font color=\"blue\">ushort</font>" },
				{ "System.Int32",	"<font color=\"blue\">int</font>" },
				{ "System.UInt32",	"<font color=\"blue\">uint</font>" },
				{ "System.Int64",	"<font color=\"blue\">long</font>" },
				{ "System.UInt64",	"<font color=\"blue\">ulong</font>" },
				{ "System.Single",	"<font color=\"blue\">float</font>" },
				{ "System.Double",	"<font color=\"blue\">double</font>" },
				{ "System.Decimal",	"<font color=\"blue\">decimal</font>" },
				{ "System.Char",	"<font color=\"blue\">char</font>" },
				{ "System.Void",	"<font color=\"blue\">void</font>" }
			};

		private static string m_RootDirectory = Path.GetDirectoryName( Environment.GetCommandLineArgs()[0] );

		private const string RefString = "<font color=\"blue\">ref</font> ";

		private static int m_AliasLength = m_Aliases.GetLength( 0 );

		private static string GetPair( Type varType, string name )
		{
			string prepend = "";
			StringBuilder append = new StringBuilder();

			Type realType = varType;

			if ( varType.IsByRef )
			{
				prepend = RefString;
				realType = varType.GetElementType();
			}

			if ( realType.IsPointer )
			{
				if ( realType.IsArray )
				{
					append.Append( '*' );

					do
					{
						append.Append( '[' );

						for ( int i = 1; i < realType.GetArrayRank(); ++i )
							append.Append( ',' );

						append.Append( ']' );

						realType = realType.GetElementType();
					} while ( realType.IsArray );

					append.Append( ' ' );
				}
				else
				{
					realType = realType.GetElementType();
					append.Append( " *" );
				}
			}
			else if ( realType.IsArray )
			{
				do
				{
					append.Append( '[' );

					for ( int i = 1; i < realType.GetArrayRank(); ++i )
						append.Append( ',' );

					append.Append( ']' );

					realType = realType.GetElementType();
				} while ( realType.IsArray );

				append.Append( ' ' );
			}
			else
			{
				append.Append( ' ' );
			}

			string fullName = realType.FullName;
			string aliased = null;// = realType.Name;

			int indexOf = Array.IndexOf( m_Types, realType );

			if ( indexOf > 0 )
			{
				aliased = String.Format( "<a href=\"type_{0}.html\">{1}</a>", indexOf, realType.Name );
			}
			else
			{
				for ( int i = 0; i < m_AliasLength; ++i )
				{
					if ( m_Aliases[i, 0] == fullName )
					{
						aliased = m_Aliases[i, 1];
						break;
					}
				}

				if ( aliased == null )
					aliased = realType.Name;
			}

			return String.Concat( prepend, aliased, append, name );
		}

		private static Type[] m_Types;

		private static void EnsureDirectory( string path )
		{
			path = Path.Combine( m_RootDirectory, path );

			if ( !Directory.Exists( path ) )
				Directory.CreateDirectory( path );
		}

		private static void Document()
		{
			EnsureDirectory( "Docs/" );
			EnsureDirectory( "Docs/Namespaces/" );
			EnsureDirectory( "Docs/Types/" );

			using ( StreamWriter index = new StreamWriter( Path.Combine( Path.Combine( m_RootDirectory, "Docs/" ), "index.html" ) ) )
			{
				index.WriteLine( "<html>" );
				index.WriteLine( "   <head>" );
				index.WriteLine( "      <title>RunUO Documentation</title>" );
				index.WriteLine( "   </head>" );
				index.WriteLine( "   <body bgcolor=\"white\" style=\"font-family: Courier New\" text=\"#000000\" link=\"#000000\" vlink=\"#000000\" alink=\"#808080\">" );
				index.WriteLine( "      <h2>Namespaces</h2>" );

				Assembly asm = Assembly.GetEntryAssembly();

				Type[] allTypes = asm.GetTypes();

				m_Types = allTypes;

				Hashtable namespaces = new Hashtable();

				foreach ( Type t in allTypes )
				{
					string nspace = t.Namespace;

					if ( nspace == null )
						continue;

					ArrayList list = (ArrayList)namespaces[nspace];

					if ( list == null )
					{
						namespaces[nspace] = list = new ArrayList();
					}

					list.Add( t );
				}

				const string getString = " <font color=\"blue\">get</font>;";
				const string setString = " <font color=\"blue\">set</font>;";
				const string  inString = "<font color=\"blue\">in</font> ";
				const string outString = "<font color=\"blue\">out</font> ";

				const string virtString = "<font color=\"blue\">virtual</font> ";
				const string ctorString = "(<font color=\"blue\">ctor</font>) ";

				/*const string  classString = "<font color=\"blue\">class</font> ";
				const string  ifaceString = "<font color=\"blue\">interface</font> ";
				const string structString = "<font color=\"blue\">struct</font> ";*/

				ArrayList entries = new ArrayList( namespaces );

				entries.Sort( new NamespaceComparer() );

				int typeIndex = 0;

				for ( int i = 0; i < entries.Count; ++i )
				{
					string nspace = (string)((DictionaryEntry)entries[i]).Key;
					ArrayList types = (ArrayList)((DictionaryEntry)entries[i]).Value;

					index.Write( "      <a href=\"Namespaces/nspace_{0}.html\">", i );
					index.Write( nspace );
					index.WriteLine( "</a><br>" );

					types.Sort( new TypeComparer() );

					using ( StreamWriter nsHtml = new StreamWriter( Path.Combine( Path.Combine( m_RootDirectory, "Docs/Namespaces/" ), String.Format( "nspace_{0}.html", i ) ) ) )
					{
						nsHtml.WriteLine( "<html>" );
						nsHtml.WriteLine( "   <head>" );
						nsHtml.WriteLine( "      <title>RunUO Documentation - {0}</title>", nspace );
						nsHtml.WriteLine( "   </head>" );
						nsHtml.WriteLine( "   <body bgcolor=\"white\" style=\"font-family: Courier New\" text=\"#000000\" link=\"#000000\" vlink=\"#000000\" alink=\"#808080\">" );
						nsHtml.WriteLine( "      <h4><a href=\"../index.html\">Back to the index</a></h4>" );
						nsHtml.WriteLine( "      <h2>{0} Types</h2>", nspace );

						for ( int j = 0; j < types.Count; ++j )
						{
							Type t = (Type)types[j];

							nsHtml.Write( "      <a href=\"../Types/type_{0}.html\">", Array.IndexOf( allTypes, t ) );
							nsHtml.Write( t.Name );
							nsHtml.WriteLine( "<br>" );

							using ( StreamWriter tHtml = new StreamWriter( Path.Combine( Path.Combine( m_RootDirectory, "Docs/Types/" ), String.Format( "type_{0}.html", Array.IndexOf( allTypes, t ) ) ) ) )
							{
								tHtml.WriteLine( "<html>" );
								tHtml.WriteLine( "   <head>" );
								tHtml.WriteLine( "      <title>RunUO Documentation - {0}</title>", t.Name );
								tHtml.WriteLine( "   </head>" );
								tHtml.WriteLine( "   <body bgcolor=\"white\" style=\"font-family: Courier New\" text=\"#000000\" link=\"#000000\" vlink=\"#000000\" alink=\"#808080\">" );
								tHtml.WriteLine( "      <h4><a href=\"../Namespaces/nspace_{0}.html\">Back to {1}</a></h4>", i, nspace );

								if ( t.IsEnum )
								{
									tHtml.WriteLine( "      <h2>{0} Members</h2>", t.Name );

									string[] names = Enum.GetNames( t );

									for ( int k = 0; k < names.Length; ++k )
									{
										tHtml.Write( "      " );
										tHtml.Write( names[k] );

										if ( k < (names.Length - 1) )
											tHtml.Write( ',' );

										tHtml.WriteLine( "<br>" );
									}
								}
								else
								{
									tHtml.Write( "      <h2>" );

									if ( t.DeclaringType != null )
									{
										tHtml.Write( "(" );

										int decIndexOf = Array.IndexOf( allTypes, t.DeclaringType );

										if ( decIndexOf == -1 )
										{
											tHtml.Write( t.DeclaringType.Name );
										}
										else
										{
											tHtml.Write( "<a href=\"type_{0}.html\">", decIndexOf );
											tHtml.Write( t.DeclaringType.Name );
											tHtml.Write( "</a>" );
										}

										tHtml.Write( ") - " );
									}

									/*if ( t.IsClass )
										tHtml.Write( classString );
									else if ( t.IsInterface )
										tHtml.Write( ifaceString );
									else if ( t.IsValueType )
										tHtml.Write( structString );*/

									tHtml.Write( t.Name );

									Type[] ifaces = t.GetInterfaces();

									if ( t.BaseType != null && t.BaseType != typeof( object ) && t.BaseType != typeof( ValueType ) && !t.BaseType.IsPrimitive )
									{
										tHtml.Write( " : " );

										int baseIndexOf = Array.IndexOf( allTypes, t.BaseType );

										if ( baseIndexOf == -1 )
										{
											tHtml.Write( t.BaseType.Name );
										}
										else
										{
											tHtml.Write( "<a href=\"type_{0}.html\">", baseIndexOf );
											tHtml.Write( t.BaseType.Name );
											tHtml.Write( "</a>" );
										}

										for ( int k = 0; k < ifaces.Length; ++k )
										{
											tHtml.Write( ", " );

											int indexOf = Array.IndexOf( allTypes, ifaces[k] );

											if ( indexOf == -1 )
											{
												tHtml.Write( ifaces[k].Name );
											}
											else
											{
												tHtml.Write( "<a href=\"type_{0}.html\">", indexOf );
												tHtml.Write( ifaces[k].Name );
												tHtml.Write( "</a>" );
											}
										}
									}
									else if ( ifaces.Length > 0 )
									{
										tHtml.Write( " : " );

										for ( int k = 0; k < ifaces.Length; ++k )
										{
											if ( k > 0 )
												tHtml.Write( ", " );

											int indexOf = Array.IndexOf( allTypes, ifaces[k] );

											if ( indexOf == -1 )
											{
												tHtml.Write( ifaces[k].Name );
											}
											else
											{
												tHtml.Write( "<a href=\"type_{0}.html\">", indexOf );
												tHtml.Write( ifaces[k].Name );
												tHtml.Write( "</a>" );
											}
										}
									}

									tHtml.WriteLine( "</h2>" );

									bool hasSubclasses = false;

									for ( int subIndex = 0; subIndex < allTypes.Length; ++subIndex )
									{
										Type subType = allTypes[subIndex];

										if ( subType.BaseType == t )
										{
											if ( !hasSubclasses )
											{
												hasSubclasses = true;

												tHtml.Write( "      <h4>Derived Types: " );
											}
											else
											{
												tHtml.Write( ", " );
											}

											tHtml.Write( "<a href=\"type_{0}.html\">", subIndex );
											tHtml.Write( subType.Name );
											tHtml.Write( "</a>" );
										}
									}

									if ( hasSubclasses )
										tHtml.WriteLine( "</h4>" );

									MemberInfo[] membs = t.GetMembers( BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Static | BindingFlags.Instance | BindingFlags.DeclaredOnly );

									Array.Sort( membs, new MemberComparer() );

									foreach ( MemberInfo mi in membs )
									{
										if ( mi is PropertyInfo )
										{
											PropertyInfo prop = (PropertyInfo)mi;

											tHtml.Write( "      " );
											tHtml.Write( GetPair( prop.PropertyType, prop.Name ) );
											tHtml.Write( '{' );

											if ( prop.CanRead )
												tHtml.Write( getString );

											if ( prop.CanWrite )
												tHtml.Write( setString );

											tHtml.WriteLine( " }<br>" );
										}
										else if ( mi is ConstructorInfo )
										{
											ConstructorInfo ctor = (ConstructorInfo)mi;

											if ( ctor.IsStatic )
												continue;

											tHtml.Write( "      " );
											tHtml.Write( ctorString );
											tHtml.Write( t.Name );
											tHtml.Write( '(' );

											ParameterInfo[] parms = ctor.GetParameters();

											if ( parms.Length > 0 )
											{
												tHtml.Write( ' ' );

												for ( int k = 0; k < parms.Length; ++k )
												{
													ParameterInfo pi = parms[k];

													if ( k != 0 )
														tHtml.Write( ", " );

													if ( pi.IsIn )
														tHtml.Write( inString );
													else if ( pi.IsOut )
														tHtml.Write( outString );

													tHtml.Write( GetPair( pi.ParameterType, pi.Name ) );
												}

												tHtml.Write( ' ' );
											}

											tHtml.WriteLine( ")<br>" );
										}
										else if ( mi is MethodInfo )
										{
											MethodInfo method = (MethodInfo)mi;

											if ( method.IsSpecialName )
												continue;

											tHtml.Write( "      " );

											if ( method.IsVirtual )
												tHtml.Write( virtString );

											tHtml.Write( GetPair( method.ReturnType, method.Name ) );
											tHtml.Write( '(' );

											ParameterInfo[] parms = method.GetParameters();

											if ( parms.Length > 0 )
											{
												tHtml.Write( ' ' );

												for ( int k = 0; k < parms.Length; ++k )
												{
													ParameterInfo pi = parms[k];

													if ( k != 0 )
														tHtml.Write( ", " );

													if ( pi.IsIn )
														tHtml.Write( inString );
													else if ( pi.IsOut )
														tHtml.Write( outString );

													tHtml.Write( GetPair( pi.ParameterType, pi.Name ) );
												}

												tHtml.Write( ' ' );
											}

											tHtml.WriteLine( ")<br>" );
										}
									}
								}

								tHtml.WriteLine( "   </body>" );
								tHtml.WriteLine( "</html>" );
							}

							++typeIndex;
						}

						nsHtml.WriteLine( "   </body>" );
						nsHtml.WriteLine( "</html>" );
					}
				}

				index.WriteLine( "   </body>" );
				index.WriteLine( "</html>" );
			}
		}
	}
}